import asyncio
import os
import orjson as json
import random
import string
from http import HTTPStatus
from typing import Union, Any, Dict, Optional, Tuple
from .constants import DEFAULT, GET, POST, PUT, DELETE, PATCH, WILDCARD
from .mocks import _listify_headers, MockRequest, MockResponse, MockContext
from .request import Request
from .response import Response
from .context import Context

class Earth:
    def __init__(self, app):
        self._app = app
        self._session_headers = {}
        self._session_cookies = {}
        self._track_session = False
        self._swaps = {}
        self._bypasses = set()

    def bypass(self, middleware):
        """Register a middleware to be bypassed during tests"""
        self._bypasses.add(middleware)

    def swap(self, old_func, new_func):
        """Register a hook to be swapped during tests"""
        self._swaps[old_func] = new_func

    # Atomic factories
    def request(self, url: str, method: str = 'GET', body: Any = b'', headers: Dict = None, **kwargs) -> Request:
        h = {**self._session_headers, **(headers or {})}
        return MockRequest(url=url, method=method, body=body, headers=h, **kwargs)

    def response(self, req: Request = None) -> Response:
        ctx = Context(self._app)
        return Response(self._app, ctx, req or self.request('/'))

    def context(self) -> Context:
        return Context(self._app)

    def trio(self, url: str = '/', method: str = 'GET', **kwargs) -> Tuple[Request, Response, Context]:
        req = self.request(url, method, **kwargs)
        ctx = self.context()
        res = Response(self._app, ctx, req)
        return req, res, ctx

    # Short synonyms
    def req(self, *args, **kwargs): return self.request(*args, **kwargs)
    def res(self, *args, **kwargs): return self.response(*args, **kwargs)
    def ctx(self, *args, **kwargs): return self.context(*args, **kwargs)

    async def _simulate(self, method: str, url: str, body: Any = b'', headers: Dict = None, subdomain: str = DEFAULT) -> Tuple[Request, Response, Context]:
        final_headers = {**self._session_headers, **(headers or {})}
        
        if isinstance(body, (dict, list)):
            body = json.dumps(body)
            final_headers.setdefault('content-type', 'application/json')
        elif isinstance(body, str):
            body = body.encode()
        
        # Prepare ASGI scope
        scope = {
            'type': 'http',
            'method': method,
            'path': url,
            'raw_path': url,
            'query_string': b'', # Improved below if needed
            'headers': _listify_headers(final_headers),
            'client': ('127.0.0.1', 8080),
            'scheme': 'http',
        }

        # Handle query string in URL
        if '?' in url:
            path, qs = url.split('?', 1)
            scope['path'] = path
            scope['raw_path'] = path
            scope['query_string'] = qs.encode()

        async def receive():
            return {'type': 'http.request', 'body': body, 'more_body': False}

        async def send(msg): pass

        engine = self._app.subdomains.get(subdomain)
        from .utils import preprocessor
        metadata = subdomain, preprocessor(scope)[1] # Update subdomain in metadata
        
        if not self._app._baked: self._app._bake_schemas()

        res = await engine.handle(scope, receive, send, metadata=metadata, application=self._app)
        
        # Consume streaming body for easy testing
        if hasattr(res.body, '__aiter__'):
            chunks = []
            async for chunk in res.body:
                chunks.append(chunk)
            res.body = b"".join(chunks)

        # Session tracking (Simplified)
        if self._track_session:
            for k, v in res.headers:
                if k.decode().lower() == 'set-cookie':
                    # This is naive but works for simple testing
                    cookie_part = v.decode().split(';')[0]
                    if '=' in cookie_part:
                        name, val = cookie_part.split('=', 1)
                        self._session_cookies[name] = val
                        # Update session headers with Cookie header for next request
                        cookie_str = "; ".join([f"{n}={v}" for n, v in self._session_cookies.items()])
                        self._session_headers['cookie'] = cookie_str

        return res._req, res, res._ctx

    async def upload(self, url: str, files: Dict[str, Any] = None, data: Dict[str, str] = None, subdomain: str = DEFAULT, **kwargs) -> Tuple[Request, Response, Context]:
        """Helper to simulate multipart/form-data uploads"""
        boundary = '----HeavenEarthBoundary' + ''.join(random.choices(string.ascii_letters + string.digits, k=16))
        body = b''
        
        # Add data fields
        if data:
            for name, value in data.items():
                body += f'--{boundary}\r\n'.encode()
                body += f'Content-Disposition: form-data; name="{name}"\r\n\r\n'.encode()
                body += f'{value}\r\n'.encode()
        
        # Add files
        if files:
            for name, content in files.items():
                filename = 'test_file'
                if isinstance(content, tuple):
                    filename, content = content
                body += f'--{boundary}\r\n'.encode()
                body += f'Content-Disposition: form-data; name="{name}"; filename="{filename}"\r\n'.encode()
                body += b'Content-Type: application/octet-stream\r\n\r\n'
                body += content + b'\r\n'
        
        if body:
            body += f'--{boundary}--\r\n'.encode()
        
        headers = kwargs.get('headers', {})
        headers['content-type'] = f'multipart/form-data; boundary={boundary}'
        kwargs['headers'] = headers
        
        return await self._simulate(POST, url, body=body, subdomain=subdomain, **kwargs)

    def SOCKET(self, url: str, subdomain: str = DEFAULT) -> 'MockSocket':
        """Helper to simulate WebSocket connections"""
        return MockSocket(self, url, subdomain=subdomain)

    async def GET(self, url, subdomain: str = DEFAULT, **kwargs): return await self._simulate(GET, url, subdomain=subdomain, **kwargs)
    async def POST(self, url, subdomain: str = DEFAULT, **kwargs): return await self._simulate(POST, url, subdomain=subdomain, **kwargs)
    async def PUT(self, url, subdomain: str = DEFAULT, **kwargs): return await self._simulate(PUT, url, subdomain=subdomain, **kwargs)
    async def DELETE(self, url, subdomain: str = DEFAULT, **kwargs): return await self._simulate(DELETE, url, subdomain=subdomain, **kwargs)
    async def PATCH(self, url, subdomain: str = DEFAULT, **kwargs): return await self._simulate(PATCH, url, subdomain=subdomain, **kwargs)

    def test(self, track_session=True):
        return EarthContextManager(self, track_session)

class EarthContextManager:
    def __init__(self, earth: Earth, track_session: bool):
        self.earth = earth
        self.track_session = track_session
        self._original_initializers = []
        self._original_deinitializers = []

    async def __aenter__(self):
        self.earth._track_session = self.track_session
        
        # Intercept hooks
        if self.earth._swaps:
            from .router import _string_to_function_handler
            import functools

            def wrap_hook(hook, swaps):
                # Unwrap our own closure if it exists to get the original func
                original = getattr(hook, '__wrapped__', hook)
                replacement = swaps.get(original)
                if replacement:
                    if isinstance(replacement, str): replacement = _string_to_function_handler(replacement)
                    
                    @functools.wraps(replacement)
                    async def hidden(app):
                        from inspect import iscoroutinefunction
                        if iscoroutinefunction(replacement): return await replacement(app)
                        else: return replacement(app)
                    return hidden
                return hook

            # Backup and swap
            self._original_initializers = list(self.earth._app.initializers)
            self._original_deinitializers = list(self.earth._app.deinitializers)
            
            self.earth._app.initializers.clear()
            self.earth._app.initializers.extend([wrap_hook(h, self.earth._swaps) for h in self._original_initializers])
            
            self.earth._app.deinitializers.clear()
            self.earth._app.deinitializers.extend([wrap_hook(h, self.earth._swaps) for h in self._original_deinitializers])

        await self.earth._app._register()
        return self.earth

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        # Restore original hooks
        if self._original_initializers:
            self.earth._app.initializers.clear()
            self.earth._app.initializers.extend(self._original_initializers)
        if self._original_deinitializers:
            self.earth._app.deinitializers.clear()
            self.earth._app.deinitializers.extend(self._original_deinitializers)
            
        await self.earth._app._unregister()

class MockSocket:
    def __init__(self, earth: Earth, url: str, subdomain: str = DEFAULT):
        self.earth = earth
        self.url = url
        self.subdomain = subdomain
        self.incoming = asyncio.Queue()
        self.outgoing = asyncio.Queue()
        self.closed = False
        self.accepted = False
        self.task = None

    async def connect(self):
        scope = {
            'type': 'websocket',
            'path': self.url,
            'raw_path': self.url,
            'query_string': b'',
            'headers': _listify_headers(self.earth._session_headers),
            'subdomain': self.subdomain,
            'client': ('127.0.0.1', 8080),
            'scheme': 'ws',
        }
        
        async def receive():
            # Initial connect or subsequent messages
            if not self.accepted:
                return {'type': 'websocket.connect'}
            return await self.incoming.get()

        async def send(msg):
            if msg['type'] == 'websocket.accept':
                self.accepted = True
            elif msg['type'] == 'websocket.send':
                await self.outgoing.put(msg.get('text') or msg.get('bytes'))
            elif msg['type'] == 'websocket.close':
                self.closed = True
                self.accepted = False

        engine = self.earth._app.subdomains.get(self.subdomain)
        if not engine:
            wildcard_engine = self.earth._app.subdomains.get(WILDCARD)
            engine = wildcard_engine if wildcard_engine else self.earth._app.subdomains.get(DEFAULT)
        
        # Handle the websocket connection in a task
        metadata = (self.subdomain, self.earth._session_headers)
        self.task = asyncio.create_task(engine.handle(scope, receive, send, metadata=metadata, application=self.earth._app))
        
        # Wait a bit for accept
        timeout = 2.0
        start = asyncio.get_event_loop().time()
        while not self.accepted and not self.closed:
            await asyncio.sleep(0.01)
            if asyncio.get_event_loop().time() - start > timeout:
                raise TimeoutError("WebSocket connection timed out")
        
        return self

    async def send(self, data):
        """Send message from client to app"""
        msg = {'type': 'websocket.receive'}
        if isinstance(data, str): msg['text'] = data
        else: msg['bytes'] = data
        await self.incoming.put(msg)

    async def receive(self, timeout=None):
        """Receive message from app to client"""
        if timeout:
            return await asyncio.wait_for(self.outgoing.get(), timeout)
        return await self.outgoing.get()

    async def close(self, code=1000):
        await self.incoming.put({'type': 'websocket.disconnect', 'code': code})
        self.closed = True
        self.accepted = False
        if self.task:
            try: await asyncio.wait_for(self.task, timeout=1.0)
            except asyncio.TimeoutError: pass



import hmac
import hashlib
import base64
import time
import struct
import typing
from secrets import compare_digest
from typing import Union, Optional
from orjson import dumps, loads, JSONDecodeError
from pytastic import Pytastic
from pytastic.exceptions import ValidationError

# --- Custom Exceptions ---
class SecurityError(Exception):
    """Base exception for all security related errors."""
    pass

class BadSignature(SecurityError):
    """Raised when the signature does not match the data."""
    pass

class SignatureExpired(BadSignature):
    """Raised when the signature is valid but the token has expired."""
    def __init__(self, message, payload=None, date_signed=None):
        super().__init__(message)
        self.payload = payload
        self.date_signed = date_signed

# --- The Core Class ---
class SecureSerializer:
    def __init__(
        self, 
        secret_keys: typing.Union[str, typing.List[str]], 
        salt: str = "app-context", 
        digest_method=hashlib.sha256
    ):
        if isinstance(secret_keys, str):
            self.secret_keys = [secret_keys.encode('utf-8')]
        else:
            self.secret_keys = [k.encode('utf-8') for k in secret_keys]
            
        self.salt = salt.encode('utf-8')
        self.digest_method = digest_method
        
    def _derive_key(self, secret_key: bytes) -> bytes:
        return hmac.new(secret_key, self.salt, self.digest_method).digest()

    def _base64_encode(self, data: bytes) -> bytes:
        return base64.urlsafe_b64encode(data).rstrip(b'=')

    def _base64_decode(self, data: bytes) -> bytes:
        pad = b'=' * (-len(data) % 4)
        return base64.urlsafe_b64decode(data + pad)

    def _sign(self, value: bytes, key: bytes) -> bytes:
        sig = hmac.new(key, value, self.digest_method).digest()
        return self._base64_encode(sig)

    def dumps(self, obj: typing.Any) -> str:
        """
        Serializes using json, signs, and timestamps.
        """
        # 1. Serialize Data
        try: json_bytes = dumps(obj)
        except TypeError as e: raise SecurityError(f"Serialization failed: {e}")

        # 2. Create Timestamp
        timestamp = int(time.time())
        ts_bytes = struct.pack('>I', timestamp)
        
        # 3. Construct Payload
        # Structure: Base64(JsonBytes).Base64(TimestampBytes)
        payload = self._base64_encode(json_bytes) + b'.' + self._base64_encode(ts_bytes)
        
        # 4. Sign with PRIMARY key
        derived_key = self._derive_key(self.secret_keys[0])
        signature = self._sign(payload, derived_key)
        
        return (payload + b'.' + signature).decode('utf-8')

    def loads(self, token: typing.Union[str, bytes], max_age: int = None, type: typing.Type = None) -> typing.Any:
        """
        Verifies token and decodes using json.
        
        :param type: Optional Type for Pytastic validation.
        """
        if isinstance(token, str):
            token = token.encode('utf-8')

        try:
            payload, signature = token.rsplit(b'.', 1)
            b64_data, b64_ts = payload.split(b'.', 1)
        except ValueError:
            raise BadSignature("Invalid token format")

        # --- KEY ROTATION CHECK ---
        client_sig_valid = False
        for secret in self.secret_keys:
            derived_key = self._derive_key(secret)
            expected_sig = self._sign(payload, derived_key)
            if compare_digest(signature, expected_sig):
                client_sig_valid = True
                break
        
        if not client_sig_valid:
            raise BadSignature("Signature mismatch")

        # --- EXPIRATION CHECK ---
        try:
            ts_bytes = self._base64_decode(b64_ts)
            timestamp = struct.unpack('>I', ts_bytes)[0]
        except Exception:
            raise BadSignature("Invalid timestamp format")

        if max_age is not None:
            age = time.time() - timestamp
            if age > max_age:
                # Try to decode mainly for the error message context
                try:
                    data = loads(self._base64_decode(b64_data).decode('utf-8'))
                except:
                    data = None
                raise SignatureExpired(f"Token expired {age} seconds ago", payload=data, date_signed=timestamp)

        # --- DECODE DATA ---
        try:
            raw_json = self._base64_decode(b64_data).decode('utf-8')
            data = loads(raw_json)
            
            if type:
                # Strict Schema Validation if type is provided
                return Pytastic().validate(type, data)
            else:
                # Generic decoding (returns dict/list)
                return data
        except JSONDecodeError as e:
            raise BadSignature(f"Payload JSON corrupted: {e}")
        except ValidationError as e:
            raise BadSignature(f"Validation failed: {e}")



from .router import App, Application
from .router import Router
from .router import Server
from .request import Request
from .response import Response
from .schema import Pytastic, PytasticError, ValidationError
from .context import Context, Key


__version__ = '1.3.10'



from ipaddress import ip_address

from .constants import DEFAULT


b_or_s = lambda x: x.decode() if isinstance(x, bytes) else x

    
def preprocessor(scope):
    headers = {}
    for header in scope.get('headers'):
        key, value = [b_or_s(e) for e in header]
        exists = headers.get(key)
        if exists:
            if isinstance(exists, list): exists.append(value)
            else: exists = [exists, value]
        else: exists = value
        headers[key] = exists

    host: bytes = headers.get('host')
    if not host: return DEFAULT, headers
    if host.startswith('http://'): host = host.replace('http://', '')
    else: host = host.replace('https://', '')
    host = host.rsplit(':')[0]
    try: ip_address(host)
    except: pass
    else: return DEFAULT, headers
    parts = host.split('.', 2)
    has_subdomain = len(parts) > 2
    return (parts[0], headers,) if has_subdomain else (DEFAULT, headers,)


class Lookup(object):
    def __init__(self, data: dict):
        self._data = data
    
    def __getattr__(self, key: str):
        value = self._data.get(key)
        if isinstance(value, dict):
            return Lookup(value)
        return value
